------------------------------------------------------------------------
--      Erasure Procedure
--     by Mara Alpuente, Santiago Escobar and Salvador Lucas
--    Version of 16/11/2001 (Universidad Politcnica de Valencia)
------------------------------------------------------------------------

---------------------------------
-- Erasure of redundant arguments
----------------------------------

--module erasureRedArgs(eraseFile,eraseProg,SyntacticErasure,is) where

import Flat
import FlatTools
import interpreter
import genint
import List

-- DataType for syntatic erasures (Arguments start with 0)
data SyntacticErasure = EFunc String [Int]   

-- DataType for substitutions
data Subs x = Subs Int x

-- Help functions to show Flat Description of a program
showFile file = do
  prog <- readFlatCurry file
  putStrLn (showFlatProg prog)

--showProg (Prog _ _ _ FuncDecls _ _) =
--  putStrLn ("["++ (flatShowListElems flatShowFunc FuncDecls) ++ "]")

-- Syntactic Erasure of a File
eraseFile erasure file =  
 if erasure==[] then putStrLn "No redundant arguments detected."
 else do 
  putStrLn ">>>Starting erasure procedure...\n"
  putStrLn (">>>Reading and compiling program "++file++"\n")
  prog <- readFlatCurry file
  let progE = eraseProg erasure prog
      fileE = file++"_erased"
      progTE = removeDupProg (removeTrivialProg progE)
      fileTE = fileE++"_trivial"
      progRE = removeDupProg (normalizeProg progTE)
      fileRE = fileTE++"_reduced"
  writeFLC (fileE++".flc") progE
  putStrLn ("\n>>>>>Saving file "++fileE++"\n")
  flc2curry fileE
  putStrLn "\n>>>Starting procedure for removing trivial rules...\n"
  writeFLC (fileTE++".flc") progTE
  putStrLn ("\n>>>>>Saving file "++fileTE++"\n")
  flc2curry fileTE
  putStrLn "\nDo you want to normalize rhs's?"
  putStrLn "Note: the TRS without trivial rules must be terminating!!!!"
  putStr "(y/n)?"
  YN <- getLine 
  if YN=="n" then done else do 
    putStrLn ">>>Starting normalizing rhs's...\n"
    writeFLC (fileRE++".flc") progRE
    putStrLn (">>>>>Saving file "++fileRE++"\n")
    flc2curry fileRE

-- Syntactic Erasure of a Program
eraseProg erasure (Prog Mod Imp TypeD FuncD OpD Tran) =
                   Prog Mod Imp TypeD FuncD2 OpD Tran
 where
   FuncD2 = map (foldFuncs (map (eraseSyn TypeD) erasure)) FuncD

-- Syntactic Erasure of a Signature & Program
eraseSyn DataTypes (EFunc EF EArgs) (Func F FArity FType FRule) =
       if EF == F then Func F FArity2 FType2 FRule2
                  else Func F FArity  FType  FRule2
 where
  FArity2 = FArity - (length EArgs);
  FType2 = eraseSynType 1 EArgs FType;
  FRule2 = eraseSynRule (EFunc EF EArgs) FRule;
  eraseSynRule _ (External Str)   = (External Str)
  eraseSynRule Era (Rule Vars Expr) = 
     if EF == F then (Rule Vars2 (eraseSynExp Era Expr3))
                else (Rule Vars  (eraseSynExp Era Expr))
   where 
    Vars2 = remove EArgs Vars;
    VarsToErase = collectVarsEra (insF F EArgs) Expr;
    Subst = buildSubsts VarsToErase F FType DataTypes;
    insF _    [] = [];
    insF FSym (x:xs) = (FSym,x,x-1):(insF FSym xs);
    Expr2 = repairTreeAfterEra VarsToErase Expr;
    Expr3 = foldFuncs (map applySubs Subst) Expr2;;
  eraseSynType _ _     (TVar TVarIndex)         = (TVar TVarIndex)
  eraseSynType _ _     (TCons Str TypeExprs) = (TCons Str TypeExprs)
  eraseSynType I Eargs (FuncType TypeEH TypeET) = 
     if I `elem` Eargs then TypeET2 else (FuncType TypeEH TypeET2)
   where TypeET2 = eraseSynType (I+1) Eargs TypeET


--eraseProgPartial [(EFunc "ej_f" [1])] (getProgram "ej")



-- Partial Erasure of a Program
--  applying substitution to rhs's, no removing arguments
eraseProgPartial erasure (Prog Mod Imp TypeD FuncD OpD Tran) =
                          Prog Mod Imp TypeD FuncD2 OpD Tran
 where
   call (EFunc F A)=eraseSynPartial (getExps F A) F A
   FuncD2 = map (foldFuncs (map call erasure)) FuncD;
   getExps EF EArgs = expEArgs
    where 
     Func _ _ FType _ = funcDecl EF FuncD;
     typeEArgs = map (\x->x FType) (map getArg EArgs);
     expEArgs = map (\x->findConstant x TypeD) typeEArgs;


-- Syntactic Erasure of a Signature & Program
eraseSynPartial Exs F1 Eargs (Func F2 FArity FType FRule) =
                              Func F2 FArity FType FRule2
 where
  FRule2 = eraseSynRule F1 Eargs Exs FRule;
  eraseSynRule _ _ _ (External Str)   = (External Str);
  eraseSynRule EF EArgs EExps (Rule Vs Ex) = Rule Vs Ex_
   where 
    Ex_ = eraseExp EF EArgs EExps Ex
    eraseExp _ _ _ (Var VarIndex) = (Var VarIndex)
    eraseExp _ _ _ (Lit Literal) = (Lit Literal)
    eraseExp F A Es (Apply Expr1 Expr2) = Apply (e Expr1) (e Expr2)
     where e x = eraseExp F A Es x
    eraseExp F A Es (Constr Vars Expr) = Constr Vars (e Expr)
     where e x = eraseExp F A Es x
    eraseExp F A Es (Or Expr1 Expr2) = Or (e Expr1) (e Expr2)
     where e x = eraseExp F A Es x
    eraseExp F A Es (Case CT Expr BranchExprs) = 
                 Case CT (e Expr) (map eB BranchExprs)
     where e x = eraseExp F A Es x
           eB (Branch P E) = Branch P (e E)
    eraseExp F A Es (Choice Expr) = Choice (e Expr)
     where e x = eraseExp F A Es x
    eraseExp F A Es (GuardedExpr Vars Expr1 Expr2) = 
                 GuardedExpr Vars (e Expr1) (e Expr2)
     where e x = eraseExp F A Es x
    eraseExp F A Es (Comb CT Str Exprs) 
     | CT == FuncCall && F == Str = Comb CT Str (map e (replace 1 A Es Exprs))
     | otherwise = Comb CT Str (map e Exprs)
     where e x = eraseExp F A Es x
    replace _ _ _  []     = []
    replace N A Es (x:xs) = replaceAux N A Es x:replace (N+1) A Es xs
    replaceAux _ []     []     x = x
    replaceAux N (n:ns) (e:es) x = if N==n then e else replaceAux N ns es x

-- Collect Vars in rhs's affected by erasure
--   EArgs is a list of triples (FuncSymbol,Argument,Variable)
--   to build a correct substitution
--   Returns a list of same kind of triples
--   It includes also original variables
collectVarsEra EArgs Exp
  | is (\(Case _ _ _)->True) Exp  = cveCase EArgs Exp
  | otherwise                     = EArgs
 where
  cveCase Vs (Case _ (Var V) BranchExprs) = cveBranchs Vs2 BranchExprs
   where 
    NewVars = cveVarsBranchs [] BranchExprs;
    Vs2 = replaceElemList V NewVars Vs;
    cveVarsBranchs Vars [] = Vars;
    cveVarsBranchs Vars (x:xs) = cveVarsBranchs (cveVarsBranch Vars x) xs;
    cveVarsBranch Vars1 (Branch (Pattern F Vars) _) = Vars1 ++ ProcessVars
     where ProcessVars = insFVars F Vars;
           insFVars F1 Vars2 = insFVars2 1 F1 Vars2
            where 
             insFVars2 _ _    [] = [];
             insFVars2 N FSym (x:xs) = (FSym,N,x):(insFVars2 (N+1) FSym xs);;
    cveVarsBranch Vars1 (Branch (LPattern _) _) = Vars1;
    replaceElemList _ _ [] = [];
    replaceElemList x L ((S,A,y):ys)=
           if x==y then (S,A,y):L++ys else (S,A,y):replaceElemList x L ys;
  cveBranchs Vars [] = Vars;
  cveBranchs Vars (x:xs) = cveBranchs (cveBranch Vars x) xs
   where
    cveBranch V (Branch _ E) = collectVarsEra V E;

-- Builds substitution for variables
--  Takes a list of triples (FSymbol,Argument,Variable)
buildSubsts [] _ _ _ = []
buildSubsts ((FSym,A,V):xs) F FType DataTypes = 
            (Subs V Exp):(buildSubsts xs F FType DataTypes)
 where
  TypeOfExp = if FSym==F then (getArg A FType) 
                         else (getD A FSym DataTypes);
  getD Arg CSym DTypes  --TODO more primitive types!!!!!!
   | CSym == ":"       = if Arg==1 then (TVar 0) else (TCons "[]" [(TVar 0)])
   | CSym == "(,...,)" = getArg 1 (TVar 0)
   | otherwise         = getD2 Arg CSym DTypes
  getD2 A1 CSym (Type _ _ ConsDecls:ts)
   | otherwise         = getDL A1 CSym ConsDecls ts
   where 
    getDL Arg C [] zs = getD2 Arg C zs;
    getDL Arg C1 ((Cons C2 _ TypeE):cs) zs = 
       if C1==C2 then getArg 1 (getN Arg TypeE) else getDL Arg C1 cs zs;;
  Exp = findConstant TypeOfExp DataTypes;

-- Get Type from Function Definition
getArg _ (TVar VarIndex) = TVar VarIndex
getArg _ (TCons Str Type2) = TCons Str Type2
getArg N (FuncType TypeEH TypeET) = if N<1 then error "getArg N<1"
                                    else if N==1 then getArg 1 TypeEH 
                                                 else getArg (N-1) TypeET

-- Find a constant in a data type
findConstant (TVar _) DType =  -- We select an arbitrary type!!!!
   findConstant (TCons "Bool" []) DType
findConstant (TCons TypeOf Type2) DType   -- TODO: more primitive Types!!!!!!!
  | TypeOf=="Bool" = Comb ConsCall "True" []
  | TypeOf=="Int"  = Lit (Intc 0)
  | TypeOf=="Char" = Lit (Charc 'a')
  | TypeOf=="[]"   = Comb ConsCall "[]" []
  | TypeOf=="()"   = Comb ConsCall "()" []
  | otherwise      = findConstantAux (TCons TypeOf Type2) DType
 where
  findConstantAux (TCons T1 T2) (Type TStr _ ConsDecls:xs) 
    | T1==TStr  = getConstant T1 ConsDecls
    | otherwise = findConstantAux (TCons T1 T2) xs
  getConstant typeof [] = error ("Constant not found in Type "++typeof)
  getConstant typeof (Cons CSym _ TEs:xs) = 
   if (all (notRecurCons typeof) TEs) 
   then Comb ConsCall CSym (map (\x->findConstant x DType) TEs) 
   else getConstant typeof xs
  notRecurCons _      (TVar _) = True
  notRecurCons typeof (FuncType TE1 TE2) = all (notRecurCons typeof) [TE1,TE2]
  notRecurCons typeof (TCons Str TEs) = 
    Str/=typeof && all (notRecurCons typeof) TEs


-- Repairs a Definitional Tree
repairTreeAfterEra Vs Exp
  | is (\(Case _ (Var _) _)->True) Exp = repairCase Vs Exp
  | otherwise                          = Exp
 where
  repairCase Vars (Case CT (Var V) BranchExprs) =
    if elemV V Vars then makeOr BranchExprs2 else Case CT (Var V) BranchExprs2
   where 
    BranchExprs2 = map repairBranch BranchExprs
    repairBranch (Branch P E) = Branch P (repairTreeAfterEra Vars E)
    elemV _ [] = False
    elemV V1 ((_,_,V2):xs) = V2==V1 || elemV V1 xs
    makeOr [(Branch _ E)] = E
    makeOr ((Branch _ E):y:ys) = Or E (makeOr (y:ys))

-- Syntactic Erasure of an Expression
eraseSynExp _   (Var VarIndex) = (Var VarIndex)
eraseSynExp _   (Lit Literal) = (Lit Literal)
eraseSynExp era (Apply Expr1 Expr2) = Apply (e Expr1) (e Expr2)
  where e x = eraseSynExp era x
eraseSynExp era (Constr Vars Expr) = Constr Vars (e Expr)
  where e x = eraseSynExp era x
eraseSynExp era (Or Expr1 Expr2) = Or (e Expr1) (e Expr2)
  where e x = eraseSynExp era x
eraseSynExp era (Case CT Expr BranchExprs) = 
                 Case CT (e Expr) (map eB BranchExprs)
  where e x = eraseSynExp era x
        eB (Branch Patt E) = Branch Patt (e E)
eraseSynExp era (Choice Expr) = Choice (e Expr)
  where e x = eraseSynExp era x
eraseSynExp era (GuardedExpr Vars Expr1 Expr2) = 
                 GuardedExpr Vars (e Expr1) (e Expr2)
  where e x = eraseSynExp era x
eraseSynExp (EFunc F EArgs) (Comb CT Str Exprs) 
  | CT == FuncCall && F == Str = Comb CT Str (map e (remove EArgs Exprs))
  | otherwise = Comb CT Str (map e Exprs)
  where 
   e x = eraseSynExp (EFunc F EArgs) x

-- Apply substitutions in an expression
applySubs (Subs V E) (Var VarIndex)=if V==VarIndex then E else (Var VarIndex)
applySubs _   (Lit Literal) = (Lit Literal)
applySubs sub (Apply Expr1 Expr2) = Apply (e Expr1) (e Expr2)
  where e x = applySubs sub x
applySubs sub (Constr Vars Expr) = Constr Vars (e Expr)
  where e x = applySubs sub x
applySubs sub (Or Expr1 Expr2) = Or (e Expr1) (e Expr2)
  where e x = applySubs sub x
applySubs sub (Case CT Expr BranchExprs) = 
               Case CT (e Expr) (map eB BranchExprs)
  where e x = applySubs sub x
        eB (Branch Patt E) = Branch Patt (e E)
applySubs sub (Choice Expr) = Choice (e Expr)
  where e x = applySubs sub x
applySubs sub (GuardedExpr Vars Expr1 Expr2) = 
                 GuardedExpr Vars (e Expr1) (e Expr2)
  where e x = applySubs sub x
applySubs sub (Comb CT Str Exprs) = Comb CT Str (map e Exprs)
  where e x = applySubs sub x

-- Compression Transformation (Reduced Erasure of a Signature & Program)
removeTrivialProg (Prog Mod Imp TypeD FuncD OpD Tran) =
                   Prog Mod Imp TypeD FuncD2 OpD Tran
 where FuncD2 = map removeTrivial FuncD

-- Removing trivial rules
--  [We assume that there exists at least one other rule apart 
--   of t -> t for each F]
removeTrivial (Func F FArity FType FRule) = (Func F FArity FType FRule2)
 where 
  FRule2 = removeT FRule
  removeT (Rule Vars Exp) = Rule Vars (removeTtree FCall Exp)
    where FCall = Comb FuncCall F (map Var Vars)
  removeT (External Str)  = External Str
  removeTtree Call Exp
   | is (\(Case _ (Var _) _)->True) Exp = removeTcase Call Exp
   | is (\(Or _ _)->True) Exp           = removeTor Call Exp
   | otherwise                          = removeTexpr Call Exp
  removeTcase Call (Case T (Var V2) BE) 
   | BE2==[]             = Var (-100)
   | otherwise           = Case T (Var V2) BE2
   where BE2 = removeTbranches Call V2 BE
  removeTor Call (Or E1 E2) 
   | E1_==Var (-100) && E2_==Var (-100) = Var (-100)
   | E1_==Var (-100)                    = E2_
   | E2_==Var (-100)                    = E1_
   | otherwise                          = Or E1_ E2_
   where E1_ = removeTtree Call E1
         E2_ = removeTtree Call E2
  removeTbranches _    _ [] = []
  removeTbranches Call V (Branch Patt E:xs) =
    removeTbranch Patt NCall E ++ removeTbranches Call V xs
   where 
    NCall = applySubs (Subs V (ePatt Patt)) Call
    ePatt (Pattern Str Vars) = Comb ConsCall Str (map Var Vars)
    ePatt (LPattern Literal) = Lit Literal
  removeTbranch Patt Call Exp = 
     if NExp == Var (-100) then [] else [Branch Patt NExp]
   where NExp = removeTtree Call Exp
  removeTexpr Call Exp = if (Call==Exp) then Var (-100) else Exp

-- Removing duplicated rules
removeDupProg (Prog Mod Imp TypeD FuncD OpD Tran) =
                   Prog Mod Imp TypeD FuncD2 OpD Tran
 where FuncD2 = map removeDupFunc FuncD

removeDupFunc (Func F FArity FType FRule) = Func F FArity FType FRule2
 where 
  FRule2 = removeD FRule
  removeD (Rule Vars Exp) = Rule Vars (removeDtree Exp)
  removeD (External Str)  = External Str
  removeDtree Exp
   | is (\(Case _ (Var _) _)->True) Exp = removeDcase Exp
   | is (\(Or _ _)->True) Exp           = removeDor Exp
   | otherwise                          = removeDexpr Exp
  removeDcase (Case T (Var V2) BE) = Case T (Var V2) BE2
   where BE2 = map removeDbranch BE
         removeDbranch (Branch Patt E) = Branch Patt (removeDtree E)
  removeDor (Or E1 E2) = if E1==E2 then E1 
                                   else Or (removeDtree E1) (removeDtree E2)
  removeDexpr Exp = Exp


-- Compression Transformation (Reduced Erasure of a Signature & Program)
normalizeProg (Prog Mod Imp TypeD FuncD OpD Tran) =
               Prog Mod Imp TypeD FuncD2 OpD Tran
 where FuncD2 = normalizeRHSs FuncD

normalizeRHSs FDs = map normalizeFunc FDs
 where
  normalizeFunc (Func F FArity FType FRule) = (Func F FArity FType FRule2)
   where FRule2 = normalizeRule FRule
  normalizeRule (Rule Vars Exp) = Rule Vars (normalizeTree Exp)
  normalizeRule (External Str)  = External Str
  -- Normalization of rhs's
  normalizeTree Exp
   | is (\(Case _ (Var _) _)->True) Exp = normalizeCase Exp
   | is (\(Or _ _)->True) Exp           = normalizeOr Exp
   | otherwise                          = normalizeExp Exp
  normalizeOr (Or e1 e2) =     -- This case might be unnecessary (TODO)
    Or (normalizeTree e1) (normalizeTree e2)
  normalizeCase (Case CT Expr Branches) = (Case CT Expr Branches2)
   where Branches2 = map normalizeBranch Branches
  normalizeBranch (Branch E1 E2) = Branch E1 (normalizeTree E2)
  normalizeExp e = normalizeConfluent FDs e
  

----------------
-- Tools
----------------

-- Get a Program
--getProgram :: String -> Prog
--getProgram file | () =:= getAux file prog = prog
--  where 
--    x,prog free
--    getAux :: String -> Prog -> IO ()
--    getAux f pr = do
--      rpr <- readFlatCurry f
--      getAux2 rpr pr
--    getAux2 rp p | rp =:= p = done

-- Get the function definitions of a Program
--getFuncs file = FuncD
--  where
--   prog = getProgram file
--   (Prog _ _ _ FuncD _ _) = prog

-- Remove a set of elements from a list
remove Es L = remove2 1 Es L
 where remove2 _ _ []     = []
       remove2 N Elems (x:xs) | N `elem` Elems = remove2 (N+1) Elems xs
                              | otherwise      = x:remove2 (N+1) Elems xs

-- functions for element N of a list
getN N (x:xs) = if N<=1 then x else getN (N-1) xs
dropN N (x:xs) = if N<=1 then xs else x:(dropN (N-1) xs)

-- Fold a list of functions to a list of elements
-- accumulating the result
foldFuncs [] y = y
foldFuncs (x:xs) y = foldFuncs xs (x y)

-- Useful for detecting patterns
is p x = findall (\_ -> p x =:= True) /= []

-- Strict
--strict x | x==x = x

-- Return rule for a function symbol
funcDecl F ((Func FCall FArity FType FRule):xs) = 
  if F==FCall then Func FCall FArity FType FRule else funcDecl F xs

--end of program
